package edu.stanford.db.rdf.syntax.generic;

import org.w3c.rdf.model.*;
import org.w3c.rdf.syntax.*;
import java.util.*;
import java.io.Writer;
import java.io.IOException;
//import org.w3c.rdf.implementation.model.*;

/**
 * RDF serializer interface.
 */

public class TripleSerializer implements RDFSerializer {

  HashMap seenModels, seenStmts;
  //  boolean firstIndent;

  /**
   * Serialize a given model into an XML character output stream.
   */
  public void serialize(Model m, Writer w) throws SerializationException, IOException, ModelException {

    //    firstIndent = true;
    seenModels = new HashMap();
    seenStmts = new HashMap();
    w.write("<?xml version=\"1.0\"?>");
    writeModel(m, w, 0);
    seenModels = null;
    w.flush();
  }

  void indent(int depth, Writer w) throws IOException {

    //    if(!firstIndent)
      w.write('\n');
      //    firstIndent = false;
    for(int i=0; i < depth; i++)
      w.write('\t');
  }

  public void writeModel(Model m, Writer w, int depth) throws ModelException, IOException {

    String id = (String)seenModels.get(m);

    if(id == null)
      seenModels.put(m, id = "m" + seenModels.size());
    else {
      w.write("<model idref=\"" + id + "\"/>");
      return;
    }

    indent(depth, w);
    w.write("<model ");
    if(depth == 0)
      w.write("xmlns=\"http://www.interdataworking.com/rdf/syntax/triples/0.1/\" ");
    w.write("id=\"" + id + "\">");

    for(Enumeration en = m.elements(); en.hasMoreElements();) {

      Statement t = (Statement)en.nextElement();
      writeStatement(t, w, depth+1, true);
      w.flush();
    }
    //    writeStatement(new StatementImpl(new ResourceImpl("subj"), new ResourceImpl("pred"), m), w, depth+1, true);
    indent(depth, w);
    w.write("</model>");
  }

  void writeEscaped(String str, Writer w) throws IOException {

    for(int i=0; i < str.length(); i++) {
      char c = str.charAt(i);
      if(c == '&')
	w.write("&amp;");
      else
	w.write(c);
    }
  }

  /**
   * ]]> cannot be there!
   * FIXME: how to we encode binary data? It is a function
   * of some more abstract layer?
   */
  void writeCDATA(String str, Writer w) throws IOException {

    int start = 0, i = 0;
    do {
      i = str.indexOf("]]>", start);
      if(i >= 0) {
	w.write(str.substring(start, i+2));
	w.write("]]><![CDATA[");
	start = i + 2;
      } else
	w.write(str.substring(start));
    } while (i >= 0 && start < str.length());
  }
  /**
   * < and &
   */
  void writeAttValue(String str, Writer w) throws IOException {

    for(int i=0; i < str.length(); i++) {
      char c = str.charAt(i);
      if(c == '<')
	w.write("&lt;");
      else if(c == '&')
	w.write("&amp;");
      else
	w.write(c);
    }
  }

  final char ABB_LONG = (char)0;
  final char ABB_CDATA = (char)1;
  final char ANYQUOTE = (char)0;

  boolean isCDATA(String s) {

    return abbrevQuote(s) == ABB_CDATA;
  }

  char abbrevQuote( String s ) {

    char quote = ANYQUOTE; // any
    boolean hasBreaks = false;
    boolean whiteSpaceOnly = true;

    for(int i=0; i < s.length(); i++) {
      char c = s.charAt(i);
      if(c == '<' /*|| c == '>'*/ || c == '&')
	return ABB_CDATA;
      else if(c == '\n')
	hasBreaks = true;

      if(c == '"' || c == '\'') {
	if(quote == ANYQUOTE)
	  quote = (c == '"') ? '\'' : '"';
	else if (c == quote)
 	  return ABB_CDATA;
      }

      if(!Character.isWhitespace(c))
	whiteSpaceOnly = false;
    }

    if(whiteSpaceOnly && hasBreaks)
      return ABB_CDATA;

    if(hasBreaks /*|| s.length() > MAX_ABBLENGTH*/) // optically nice value
      return whiteSpaceOnly ? ABB_CDATA : ABB_LONG;

    return quote == ANYQUOTE ? '"' : quote;
  }


  void writeNode(String tag, RDFNode n, Writer w, int depth) throws ModelException, IOException {

    indent(depth, w);
    w.write("<" + tag + ">");
    if(n instanceof Literal) {
      String value = ((Literal)n).getLabel();
      boolean isCDATA = isCDATA(value);
      if(isCDATA) {
	w.write("<l xml:space=\"preserve\">");
	writeCDATA(value, w);
      } else {
	w.write("<l>");
	w.write(n.getLabel());
      }
      w.write("</l>");
    } else if(n instanceof Statement)
      writeStatement((Statement)n, w, depth, false);
    else if(n instanceof Model) {
      writeModel((Model)n, w, depth);
    } else if(n instanceof Resource) {
      Resource r = (Resource)n;
      String ns = r.getNamespace();
      String name = r.getLocalName();
      if(ns != null && ns.length() > 0)
	w.write("<r name=\"" + name + "\" ns=\"" + ns + "\"/>");
      else
	w.write("<r uri=\"" + r.getURI() + "\"/>");
    } else { // resource
      w.write(n.getLabel());
    }
    w.write("</" + tag + ">");
  }

  void writeStatement(Statement st, Writer w, int depth, boolean forceBreak) throws ModelException, IOException {
    
    String id = (String)seenStmts.get(st);

    if(id == null)
      seenStmts.put(st, id = "s" + seenStmts.size());
    else {
      if(forceBreak)
	indent(depth, w);
      w.write("<stmt idref=\"" + id + "\"/>");
      return;
    }
    if(forceBreak)
      indent(depth, w);
    w.write("<stmt id=\"" + id + "\">");
    writeNode("s", st.subject(), w, depth+1);
    writeNode("p", st.predicate(), w, depth+1);
    writeNode("o", st.object(), w, depth+1);
    w.write("</stmt>");
  }
}
